import numpy as np
import dask.array as da
import skimage.exposure
import sklearn.linear_model
from loguru import logger
import tqdm.dask

from . import register
from . import block_affine
from . import img_util


def block_affine_transformed_moving_img(ref_img, moving_img, mxs):
    assert img_util.is_single_channel(ref_img)
    if img_util.is_single_channel(moving_img) and moving_img.ndim == 2:
        return da.map_blocks(
            block_affine.block_affine_dask,
            mxs,
            chunks=ref_img.chunks,
            dtype=moving_img.dtype,
            src_array=moving_img
        )
    else:
        return da.array([
            da.map_blocks(
                block_affine.block_affine_dask,
                mxs,
                chunks=ref_img.chunks,
                dtype=moving_img.dtype,
                src_array=c
            )
            for c in moving_img
        ])


def block_shifts(ref_img, moving_img, pcc_kwargs=None):
    if pcc_kwargs is None:
        pcc_kwargs = dict(sigma=0, upsample=1)
    return da.map_blocks(
        lambda a, b: np.atleast_2d(
            register.phase_cross_correlation(a, b, **pcc_kwargs)[0]
        ),
        ref_img,
        moving_img,
        dtype=np.float32
    )


def constrain_block_shifts(shifts, grid_shape):
    num_rows, num_cols = grid_shape
    distances = np.linalg.norm(shifts, axis=1)
    threshold_distance = skimage.filters.threshold_triangle(
        distances
    )

    high_confidence_blocks = distances < threshold_distance

    lr = sklearn.linear_model.LinearRegression()
    block_coords = np.mgrid[:num_rows, :num_cols].reshape(2, -1).T
    lr.fit(
        block_coords[high_confidence_blocks],
        shifts[high_confidence_blocks]
    )
    predicted_shifts = lr.predict(block_coords)
    diffs = shifts - predicted_shifts
    distance_diffs = np.linalg.norm(diffs, axis=1)
    passed = (
        distance_diffs <
        skimage.filters.threshold_triangle(distance_diffs)
    )
    fitted_shifts = shifts.copy()
    fitted_shifts[~passed] = predicted_shifts[~passed]
    return fitted_shifts


def viz_shifts(shifts, grid_shape, vmax=None):
    import matplotlib.pyplot as plt
    distances = np.linalg.norm(shifts, axis=1)
    if vmax is None:
        vmax = skimage.filters.threshold_triangle(distances)
    plt.figure()
    plt.imshow(distances.reshape(grid_shape), vmax=vmax)


def block_affine_matrices(mx, shifts):
        
    def shift_affine_mx(mx, shift):
        y, x = shift
        mx_shift = np.eye(3)
        mx_shift[:2, 2] = x, y
        return mx_shift @ mx

    mxs = [
        shift_affine_mx(mx, s)
        for s in shifts
    ]
    return np.array(mxs)


def block_affine_matrices_da(mxs, grid_shape):
    num_rows, num_cols = grid_shape
    grid = np.arange(num_rows * num_cols).reshape(grid_shape)
    mxs = np.vstack([
        np.hstack(mxs[r])
        for r in grid
    ])
    return da.from_array(mxs, chunks=3)


class Aligner:

    def __init__(
        self,
        ref_img,
        moving_img,
        ref_thumbnail,
        moving_thumbnail,
        ref_thumbnail_down_factor,
        moving_thumbnail_down_factor
    ) -> None:
        self.ref_img=ref_img
        self.moving_img=moving_img
        self.ref_thumbnail=ref_thumbnail
        self.moving_thumbnail=moving_thumbnail
        self.ref_thumbnail_down_factor=ref_thumbnail_down_factor
        self.moving_thumbnail_down_factor=moving_thumbnail_down_factor

    def coarse_register_affine(self, n_keypoints=2000):
        ref_img = self.ref_thumbnail
        moving_img = self.moving_thumbnail
        affine_matrix = register.feature_based_registration(
            ref_img, moving_img,
            n_keypoints=n_keypoints,
            plot_match_result=True
        )
        self.coarse_affine_matrix = np.vstack(
            [affine_matrix, [0, 0, 1]]
        )
    
    @property
    def affine_matrix(self):
        if not hasattr(self, 'coarse_affine_matrix'):
            self.coarse_register_affine()
        affine = skimage.transform.AffineTransform
        mx_ref = affine(scale=1/self.ref_thumbnail_down_factor).params
        mx_moving = affine(scale=1/self.moving_thumbnail_down_factor).params
        affine_matrix = (
            np.linalg.inv(mx_ref) @
            self.coarse_affine_matrix.copy() @
            mx_moving
        )
        return affine_matrix
    
    @property
    def tform(self):
        return skimage.transform.AffineTransform(
            matrix=self.affine_matrix
        )
    
    def affine_transformed_moving_img(self, mxs=None):
        if mxs is None:
            mxs = self.affine_matrix
        ref_img = self.ref_img
        moving_img = self.moving_img

        return block_affine_transformed_moving_img(
            ref_img, moving_img, mxs
        )
    
    def compute_shifts(self, pcc_kwargs=None):
        logger.info(f"Computing block-wise shifts")
        ref_img = self.ref_img
        moving_img = self.affine_transformed_moving_img(self.affine_matrix)
        shifts_da = block_shifts(ref_img, moving_img, pcc_kwargs=pcc_kwargs)
        with tqdm.dask.TqdmCallback(
            ascii=True, desc='Computing shifts',
        ):
            shifts = shifts_da.compute()
        self.shifts = shifts.reshape(-1, 2)

    @property
    def grid_shape(self):
        return self.ref_img.numblocks

    @property
    def num_blocks(self):
        return self.ref_img.npartitions

    def constrain_shifts(self):
        if not hasattr(self, 'original_shifts'):
            self.original_shifts = self.shifts.copy()
        self.shifts = constrain_block_shifts(
            self.original_shifts,
            self.grid_shape
        )
    
    @property
    def block_affine_matrices(self):
        mx = self.affine_matrix
        shifts = self.shifts
        return block_affine_matrices(mx, shifts)

    @property
    def block_affine_matrices_da(self):
        return block_affine_matrices_da(
            self.block_affine_matrices,
            self.grid_shape
        )



import numpy as np
import cv2
import skimage.transform
import skimage.filters

def cv2_pyramid(img, max_size=1024):
    size_max = np.array(img.shape).max()
    pyramid = [img]
    while size_max > max_size:
        img = cv2.pyrDown(img)
        pyramid.append(img)
        size_max //= 2
    return pyramid


def whiten(img, sigma=1):
    border_mode = cv2.BORDER_REFLECT
    g_img = img if sigma == 0 else cv2.GaussianBlur(
        img, (0, 0), sigma, 
        borderType=border_mode
    ).astype(np.float32)
    log_img = cv2.Laplacian(
        g_img, cv2.CV_32F, ksize=1,
        borderType=border_mode
    )
    # log_img = cv2.convertScaleAbs(log_img)
    return log_img


def cv2_to_uint8(img):
    return cv2.normalize(
        src=img, dst=None,
        alpha=255, beta=0,
        norm_type=cv2.NORM_MINMAX,
        dtype=cv2.CV_8U
    )


def is_single_channel(img):
    is_2d = img.ndim == 2
    is_flat_3d = img.ndim == 3 and img.shape[0] == 1
    return is_2d or is_flat_3d


def entropy_mask(img, kernel_size=9):
    img = skimage.exposure.rescale_intensity(
        img, out_range=np.uint8
    ).astype(np.uint8)
    entropy = skimage.filters.rank.entropy(img, np.ones((kernel_size, kernel_size)))
    return entropy > skimage.filters.threshold_otsu(entropy)


def is_brightfield_img(img, max_size=100):
    downscale_factor = int(max(img.shape) / max_size)
    thumbnail = skimage.transform.downscale_local_mean(img, (downscale_factor, downscale_factor))
    mask = entropy_mask(thumbnail)
    # is using mean better?
    return np.median(thumbnail[mask]) < np.median(thumbnail[~mask])


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['palom', 'palom.cli', 'palom.cli.schema']

package_data = \
{'': ['*']}

install_requires = \
['dask>=2021.10.0',
 'imagecodecs>=2021.11.11',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.4.3,<4.0.0',
 'napari-lazy-openslide>=0.2.0',
 'numpy>=1.21.3,<2.0.0',
 'opencv-python>=4.5.3.56,<5.0.0.0',
 'scikit-image>=0.18',
 'scikit-learn>=0.24',
 'scipy>=1.7.1,<2.0.0',
 'tifffile>=2021.10.12',
 'tqdm>=4.62.3,<5.0.0',
 'yamale>=4.0.2,<5.0.0',
 'zarr>=2.10.0,<3.0.0']

entry_points = \
{'console_scripts': ['palom-svs = palom.cli.svs:main',
                     'palom-svs-helper = palom.cli.helper:main']}

setup_kwargs = {
    'name': 'palom',
    'version': '2022.8.2',
    'description': 'Piecewise alignment for layers of mosaics',
    'long_description': '<h2>\n    <img alt="palom" height="60" src="palom-logo.svg">\n    <br>\n    Piecewise alignment for layers of mosaics\n</h2>\n\nPalom started as a tool for registering [whole-slide\nimages](https://en.wikipedia.org/wiki/Digital_pathology) of the same [FFPE\nsection](https://en.wikipedia.org/wiki/Histology#Sample_preparation) with\ndifferent [IHC stainings](https://en.wikipedia.org/wiki/Immunohistochemistry).\n\n---\n\n## Installation\n\nInstalling palom in a fresh conda environment is recommended. [Instruction for\ninstalling miniconda](https://docs.conda.io/en/latest/miniconda.html)\n\n### Step 1\n\nCreate a named conda environment - _palom_, in the following example, and activate the environment.\n\n```\nconda create -n palom python=3.7 pip -c conda-forge\nconda activate palom\n```\n\n### Step 2\n\nInstall openslide in the conda environment.\n\n- For MacOS, use conda-forge channel\n\n```\nconda install openslide -c conda-forge\n```\n\n- For Windows and Linux, use sdvillal channel\n\n```\nconda install openslide -c sdvillal\n```\n\n### Step 3\n\nInstall palom from pypi in the conda environment.\n\n```\npython -m pip install palom\n```\n\n---\n\n## CLI usage\n\n### Configuration YAML file\n\nPalom CLI tool merges multiple SVS files into a pyramidal OME-TIFF file, with\nthe option to perform preset stain separation (5 modes are available - `output\nmode`: `hematoxylin`, `aec`, `dab`, `grayscale`, `color`)\n\nA user-defined configuration YAML file is required for the run. A configuration\nexample can be printed to the console by running\n\n```bash\npalom-svs show example\n```\n\n```yaml\ninput dir: Y:\\user\\me\\projects\\data\\mihc\noutput full path: Y:\\user\\me\\projects\\analysis\\mihc\\2021\\skin-case-356.ome.tif\n\nreference image:\n    filename: 20210111/skin_case_356_HEM_C11R3_HEM.svs\n    output mode: hematoxylin\n    channel name: Hematoxylin\n\nmoving images:\n- filename: 20210101/skin_case_356_HEM_C01R1_PD1.svs\n  output mode: aec\n  channel name: PD-1\n- filename: 20210101/skin_case_356_HEM_C01R2_PDL1.svs\n  output mode: aec\n  channel name: PD-L1\n```\n\nTo show the configuration schema, run the following command\n\n```bash\npalom-svs show schema\n```\n\n### Use the helper script to generate the configuration file\n\n[A helper\nscript](https://github.com/Yu-AnChen/palom/blob/main/palom/cli/helper.py) is\nincluded showing how to automatically generate the configuration file if the SVS\nfiles are organized and have specific naming pattern.\n\nHere\'s an example directory containing many SVS files\n\n```\nY:\\DATA\\SARDANA\\MIHC\\768473\\RAW\n    CBB_SARDANA_768473_C04R1_CD8.svs\n    KB_SARDANA_768473_C01R1_PD1.svs\n    KB_SARDANA_768473_C01R2_PDL1.svs\n    KB_SARDANA_768473_C01R3_Hem.svs\n    KB_SARDANA_768473_C02R1_CD4.svs\n    KB_SARDANA_768473_C03R1_CD3.svs\n    KB_SARDANA_768473_C03R3_DCLAMP.svs\n```\n\nRunning the following command to generate the configuration file\n\n```bash\npalom-svs-helper -i "Y:\\DATA\\SARDANA\\MIHC\\768473\\RAW" -n "*Hem*" -o "Y:\\DATA\\SARDANA\\MIHC\\768473\\RAW\\palom\\768473.ome.tif" -c "Y:\\DATA\\SARDANA\\MIHC\\768473\\768473.yml"\n```\n\nAnd the resulting `Y:\\DATA\\SARDANA\\MIHC\\768473\\768473.yml` file \n\n```yaml\ninput dir: Y:\\DATA\\SARDANA\\MIHC\\768473\\RAW\noutput full path: Y:\\DATA\\SARDANA\\MIHC\\768473\\RAW\\palom\\768473.ome.tif\nreference image:\n  filename: .\\KB_SARDANA_768473_C01R3_Hem.svs\n  output mode: hematoxylin\n  channel name: Hem-C01R3\nmoving images:\n- filename: .\\KB_SARDANA_768473_C01R1_PD1.svs\n  output mode: aec\n  channel name: PD1-C01R1\n- filename: .\\KB_SARDANA_768473_C01R2_PDL1.svs\n  output mode: aec\n  channel name: PDL1-C01R2\n- filename: .\\KB_SARDANA_768473_C02R1_CD4.svs\n  output mode: aec\n  channel name: CD4-C02R1\n- filename: .\\KB_SARDANA_768473_C03R1_CD3.svs\n  output mode: aec\n  channel name: CD3-C03R1\n- filename: .\\KB_SARDANA_768473_C03R3_DCLAMP.svs\n  output mode: aec\n  channel name: DCLAMP-C03R3\n- filename: .\\CBB_SARDANA_768473_C04R1_CD8.svs\n  output mode: aec\n  channel name: CD8-C04R1\n```\n\nAfter reviewing the configuration file, process those SVS files by running\n\n```bash\npalom-svs run -c "Y:\\DATA\\SARDANA\\MIHC\\768473\\768473.yml"\n```\n\nWhen the process is finished, a pyramidal OME-TIFF file will be generated along\nwith PNG files showing the feature-based registration results and a log file. \n\n```\nY:\\DATA\\SARDANA\\MIHC\\768473\\RAW\n│   CBB_SARDANA_768473_C04R1_CD8.svs\n│   KB_SARDANA_768473_C01R1_PD1.svs\n│   KB_SARDANA_768473_C01R2_PDL1.svs\n│   KB_SARDANA_768473_C01R3_Hem.svs\n│   KB_SARDANA_768473_C02R1_CD4.svs\n│   KB_SARDANA_768473_C03R1_CD3.svs\n│   KB_SARDANA_768473_C03R3_DCLAMP.svs\n│\n└───palom\n    │   768473.ome.tif\n    │\n    └───qc\n            01-KB_SARDANA_768473_C01R1_PD1.svs.png\n            02-KB_SARDANA_768473_C01R2_PDL1.svs.png\n            03-KB_SARDANA_768473_C02R1_CD4.svs.png\n            04-KB_SARDANA_768473_C03R1_CD3.svs.png\n            05-KB_SARDANA_768473_C03R3_DCLAMP.svs.png\n            06-CBB_SARDANA_768473_C04R1_CD8.svs.png\n            768473.ome.tif.log\n```\n\n---\n\n## Scripting\n\n__WARNING__ API may change in the future\n\n### For SVS files\n\n```python\nimport palom\n\nc1r = palom.reader.SvsReader(r\'Y:\\DATA\\SARDANA\\MIHC\\75684\\GG_TNP_75684_D21_C11R3_HEM.svs\')\nc2r = palom.reader.SvsReader(r\'Y:\\DATA\\SARDANA\\MIHC\\75684\\GG_TNP_75684_D23_C01R1_PD1.svs\')\n\nLEVEL = 1\nTHUMBNAIL_LEVEL = 2\n\nc1rp = palom.color.PyramidHaxProcessor(c1r.pyramid, thumbnail_level=THUMBNAIL_LEVEL)\nc2rp = palom.color.PyramidHaxProcessor(c2r.pyramid, thumbnail_level=THUMBNAIL_LEVEL)\n\nc21l = palom.align.Aligner(\n    c1rp.get_processed_color(LEVEL), \n    c2rp.get_processed_color(LEVEL),\n    ref_thumbnail=c1rp.get_processed_color(THUMBNAIL_LEVEL).compute(),\n    moving_thumbnail=c2rp.get_processed_color(THUMBNAIL_LEVEL).compute(),\n    ref_thumbnail_down_factor=c1r.level_downsamples[THUMBNAIL_LEVEL] / c1r.level_downsamples[LEVEL],\n    moving_thumbnail_down_factor=c2r.level_downsamples[THUMBNAIL_LEVEL] / c2r.level_downsamples[LEVEL]\n)\n\nc21l.coarse_register_affine()\nc21l.compute_shifts()\nc21l.constrain_shifts()\n\nc21l.block_affine_matrices_da\n\nc2m = palom.align.block_affine_transformed_moving_img(\n    c1rp.get_processed_color(LEVEL),\n    c2rp.get_processed_color(LEVEL, \'aec\'),\n    mxs=c21l.block_affine_matrices_da\n)\n\npalom.pyramid.write_pyramid(\n    palom.pyramid.normalize_mosaics([c2m]),\n    r"Y:\\DATA\\SARDANA\\MIHC\\75684\\mosaic.ome.tif",\n    pixel_size=c1r.pixel_size*c1r.level_downsamples[LEVEL],\n)\n```\n\n### For TIFF and OME-TIFF files\n\n```python\nimport palom\n\n# reference image is a multichannel immunofluoroscence imaging\nc1r = palom.reader.OmePyramidReader(r"Z:\\P37_Pilot2\\P37_S12_Full.ome.tiff")\n# moving image is a brightfield imaging (H&E staining) of the same tissue\n# section as the reference image\nc2r = palom.reader.OmePyramidReader(r"Z:\\P37_Pilot2\\HE\\P37_S12_E033_93_HE.ome.tiff")\n\n# use second-to-the-bottom pyramid level for a quick test; set `LEVEL = 0` for\n# processing lowest level pyramid (full resolution)\nLEVEL = 1\n# choose thumbnail pyramid level for feature-based affine registration as\n# initial coarse alignment\n# `THUMBNAIL_LEVEL = c1r.get_thumbnail_level_of_size(2000)` might be a good\n# starting point\nTHUMBNAIL_LEVEL = 3\n\nc21l = palom.align.Aligner(\n    # use the first channel (Hoechst staining) in the reference image as the\n    # registration reference\n    ref_img=c1r.read_level_channels(LEVEL, 0),\n    # use the second channel (G channel) in the moving image, it usually has\n    # better contrast\n    moving_img=c2r.read_level_channels(LEVEL, 1),\n    # select the same channels for the thumbnail images\n    ref_thumbnail=c1r.read_level_channels(THUMBNAIL_LEVEL, 0).compute(),\n    moving_thumbnail=c2r.read_level_channels(THUMBNAIL_LEVEL, 1).compute(),\n    # specify the downsizing factors so that the affine matrix can be scaled to\n    # match the registration reference\n    ref_thumbnail_down_factor=c1r.level_downsamples[THUMBNAIL_LEVEL] / c1r.level_downsamples[LEVEL],\n    moving_thumbnail_down_factor=c2r.level_downsamples[THUMBNAIL_LEVEL] / c2r.level_downsamples[LEVEL]\n)\n\n# run feature-based affine registration using thumbnails\nc21l.coarse_register_affine(n_keypoints=4000)\n# after coarsly affine registered, run phase correlation on each of the\n# corresponding chunks (blocks/pieces) to refine translations\nc21l.compute_shifts()\n# discard incorrect shifts which is usually due to low contrast in the\n# background regions; this is needed for WSI but maybe not for ROI images\nc21l.constrain_shifts()\n\n# configure the transformation of aligning the moving image to the registration\n# reference\nc2m = palom.align.block_affine_transformed_moving_img(\n    ref_img=c1r.read_level_channels(LEVEL, 0),\n    # select all the three channels (RGB) in moving image to transform\n    moving_img=c2r.pyramid[LEVEL],\n    mxs=c21l.block_affine_matrices_da\n)\n\n# write the registered images to a pyramidal ome-tiff\npalom.pyramid.write_pyramid(\n    mosaics=palom.pyramid.normalize_mosaics([\n        # select only the first three channels in referece image to be written\n        # to the output ome-tiff; for writing all channels, use\n        # `c1r.pyramid[LEVEL]` instead\n        c1r.read_level_channels(LEVEL, [0, 1, 2]),\n        c2m\n    ]),\n    output_path=r"Z:\\P37_Pilot2\\mosaic.ome.tif",\n    pixel_size=c1r.pixel_size*c1r.level_downsamples[LEVEL]\n)\n```\n',
    'author': 'Yu-An Chen',
    'author_email': 'atwood12@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/palom/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.8,<4.0.0',
}


setup(**setup_kwargs)



from __future__ import annotations
import numpy as np
import dask.array as da
import skimage.color
import skimage.exposure

from . import extract


class HaxProcessor:

    def __init__(
        self,
        contrast_ref_img: da.Array | np.array,
        channel_axis: int = 0,
    ) -> None:
        self.contrast_ref_img = np.moveaxis(contrast_ref_img, channel_axis, 2)
        assert self.contrast_ref_img.ndim == 3, (
            f"`contrast_ref_img` must be 3D, preferably in (C, Y, X) order"
        )

    def find_processed_color_contrast_range(self, mode: str):
        supported_modes = ['grayscale', 'hematoxylin', 'aec', 'dab']
        assert mode in supported_modes
        
        mode_range = f'_{mode}_range'
        if hasattr(self, mode_range):
            return getattr(self, mode_range)
        
        contrast_ref_img = self.contrast_img
        
        process_func = self.__getattribute__(f"rgb2{mode}")
        img = process_func(contrast_ref_img)

        setattr(self, mode_range, (
            img.min(), img.max()
        ))
        return getattr(self, mode_range)

    @property
    def contrast_img(self):
        if hasattr(self, '_contrast_img'):
            return self._contrast_img
        self._contrast_img = self.contrast_ref_img
        if isinstance(self._contrast_img, da.Array):
            self._contrast_img = self._contrast_img.compute()
        return self._contrast_img

    def rgb2grayscale(self, rgb_img):
        import cv2
        return cv2.cvtColor(rgb_img, cv2.COLOR_BGR2GRAY)
        # return skimage.color.rgb2gray(rgb_img).astype(np.float32)

    def rgb2aec(self, rgb_img):
        return extract.rgb2aec(rgb_img).astype(np.float32)

    def rgb2dab(self, rgb_img):
        hdx = skimage.color.separate_stains(rgb_img, skimage.color.hdx_from_rgb)
        return hdx[..., 1].astype(np.float32)

    def rgb2hematoxylin(self, rgb_img):
        hax = skimage.color.separate_stains(rgb_img, skimage.color.hax_from_rgb)
        return hax[..., 0].astype(np.float32)

    def get_processed_color(self, rgb_img, mode='grayscale'):
        assert mode in ['grayscale', 'hematoxylin', 'aec', 'dab']

        process_func = self.__getattribute__(f"rgb2{mode}")
        intensity_range = self.find_processed_color_contrast_range(mode)
        
        processed = da.map_blocks(
            process_func,
            rgb_img,
            dtype=np.float32,
            drop_axis=2
        )

        return da.map_blocks(
            lambda x: skimage.exposure.rescale_intensity(
                x, in_range=intensity_range, out_range=(0, 1)
            ).astype(np.float32),
            processed,
            dtype=np.float32
        )


class PyramidHaxProcessor(HaxProcessor):

    def __init__(
        self,
        pyramid: list[da.Array],
        thumbnail_level: int = None
    ) -> None:
        if thumbnail_level is None:
            thumbnail_level = len(pyramid) - 1
        super().__init__(pyramid[thumbnail_level], channel_axis=0)
        self.pyramid = pyramid
    
    def get_processed_color(self, level, mode='grayscale', out_dtype=None):
        if mode == 'color':
            return self.pyramid[level]
        rgb_img = np.moveaxis(self.pyramid[level], 0, 2)
        processed = super().get_processed_color(rgb_img, mode=mode)
        if out_dtype is None:
            out_dtype = rgb_img.dtype
        if isinstance(out_dtype, np.dtype):
            out_dtype = out_dtype.type
        return da.map_blocks(
            lambda x: skimage.exposure.rescale_intensity(
                x, in_range=(0, 1), out_range=out_dtype
            ).astype(out_dtype),
            processed,
            dtype=out_dtype
        )
